// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// 
// Module: DummyPixelIterator
// Author: Hassan Siddiqui
//
// Description: 
//
// An NxM pixel iterator, returning a single value for every pixel.
//
// Modifications: See ChangeLog
// 
// $Id: DummyPixelIterator.cc,v 1.1 2012/01/25 15:46:31 rsaxton Exp $

/* General classes */
#include<cassert>

/* SAS-specific classes */
#include "Dal.h"
#include "errstr.h"
#include "Tracer.h"

/* Local classes */
#include "DummyPixelIterator.h"

/* Constructors */
DummyPixelIterator::DummyPixelIterator( 
		   		       double xmin,
				       double xmax,
				       unsigned xbins,
				       double ymin,
				       double ymax,
				       unsigned ybins):
  _firstcalled(false),
  _xbins(xbins),
  _wcsconv(0)
{ 
  PROC;
  
  errstr << "DummyPixelIterator called."
	 << Sas::message(Msg::App, Msg::Verbose);

  if (xmin >= xmax)
    {
      errstr << " X range for map ( min =" << xmin
	     << ", max ="<<xmax
	     << " ) is invalid. "
	     << Sas::error("MapXRangeInvalid");

      ALTRETURN;
    }

  if (ymin >= ymax)
    {
      errstr << " Y range for map ( min =" << ymin
	     << ", max ="<<ymax
	     << " ) is invalid. "
	     << Sas::error("MapYRangeInvalid");
    }


  _size = xbins * ybins;

  if (_size == 0)
    {
      errstr << " The number of bins requested is zero "
	     << " ( xbins =" << xbins
	     << ", ybins =" << ybins
	     << Sas::error("MapSizeInvalid");
    }

  _xdelt = (xmax-xmin)/xbins;
  _ydelt = (ymax-ymin)/ybins;
  _xbit = xmin + _xdelt/2.0 - _xdelt;
  _ybit = ymin + _ydelt/2.0 - _ydelt;

  //cout<<"BITS: "<<_xdelt<<" "<<_ydelt<<" "<<_xbit<<" "<<_ybit<<endl;

  /* Create WcsConvertor object - to pass widths */
  _wcsconv = new WCSConvertor(
			      "DETX",
			      1,  // 1st pixel in array is [1,1]
			      xmin + (_xdelt/2),
			      _xdelt,
			      "DETY",
			      1, // 1st pixel in array is [1,1]
			      ymin + (_ydelt/2),
			      _ydelt);

  ENDPROC;
}


/* Destructor */
DummyPixelIterator::~DummyPixelIterator()
{
  PROC;
    delete _wcsconv;   _wcsconv=0;
  ENDPROC;
}


/* Iterator methods */
void DummyPixelIterator::first()
{
  _firstcalled = true;
  _index=0;
  _xindex=1;
  _yindex=1;
}

void DummyPixelIterator::next()
{
  assert (!isDone());
  _index++;

  _xindex++;
  if (_xindex > _xbins)
    {
      _xindex=1;
      _yindex++;
    }
}

bool DummyPixelIterator::isDone()
{
  PROC;
  assert(_firstcalled);
  RETURN (_index>=_size);
}


PixelElement DummyPixelIterator::currentItem()
{
  /* Convert x and y positions to detector coordinates */ 

   //CoordPair physcoord = _wcsconv->physicalCoords(_xindex, _yindex);
   //double px = _xindex*_xdelt+_xbit;
   //double py = _yindex*_ydelt+_ybit;

   PixelElement result (_xindex,
			_yindex,
                        _xindex*_xdelt+_xbit,
                        _yindex*_ydelt+_ybit,
			_xdelt,
			_ydelt,
			1);

   //cout<<"DBG: "<<_xindex<<" "<<_yindex<<" "<<_xindex*_xdelt+_xbit<<" "<<_yindex*_ydelt+_ybit<<endl;

  RETURN result;
}

unsigned DummyPixelIterator::size() const
{
  PROC;
  RETURN _size;
}

unsigned DummyPixelIterator::xbins() const
{
  PROC;
  RETURN _xbins;
}

/* WCSConvertor object */
const WCSConvertor * DummyPixelIterator::wcsconvertor() const
{
  PROC;
  RETURN _wcsconv;
}

