! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! File to allow tasks to access the ssclib conversion routines
!
module ssclib_glue

  use types
  use constants ! pion180
  use coordinate
  use ssc_misc
  use wcs_aux           ! for WcsT 
  use dal
  use cal
  use cal_aux

  implicit none

! Here we declare a global WCS transformation structure to be used throughout
! and the global attitude matrices
  type(WcsT) :: wcs_image ! Transformation from RA/DEC to image pixels
  type(WcsT) :: wcs_sky ! Transformation from RA/DEC to X/Y pixels
  type(SpacecraftAttitudeType) :: scAttitude, instrumAttitude
  integer(int32) :: instrumentId
  real(double) :: timeStamp


! The procedure is to set the WCS structure by first calling "setwcs"
! and then call the transformation routines:
!
! callradectoimpixels(ra,dec,xim,yim) - converts ra/dec to image pixel
! callradectoskypixels(ra,dec,x,y) - converts ra/dec to sky (X/Y) pixel
! callxytoradec(x,y,ra,dec) - converts sky (X/Y) pixel to ra/dec
! calldettoradec(detx,dety,ra,dec) - converts det coords to ra/dec 
! calldettoxy(detx,dety,x,y) - converts det coords to sky (X/Y) 
! calldettotel(detx,dety,theta,phi) - converts det coords to tel coords
! callxytodet(x,y,detx,dety) - converts sky (X/Y) coords to det 
! callrawtodet(rawx,rawy,detx,dety) - converts raw coords to det 
! linearToImPixels(x,y,xim,yim) - converts det (or EQU) to image pixel
! imPixelsToLinear(xim,yim,x,y) - converts image pixel to linear (DET/EQU)
! callimpixelstoradec(xim,yim,ra,dec) - converts image pixel to ra/dec
!
! Care must be taken to reset WCS if the routines are used with more than 
! one set of WCS coordinates, e.g. taken from two image files
!*************************************************************
!
!*************************************************************
contains
!
! Routine to convert between ra/dec and image pixel
subroutine callradectoimpixels(ra, dec, xpixel, ypixel)

    real(double), intent(in) :: ra, dec ! Input units are radians
    real(double), intent(out) :: xpixel, ypixel
    real(double) :: detx, dety
    real(single) :: xs, ys

   ! If the transform is RA---TAN then use the ssclib routine
    select case(wcs_image%x%ctype)
    case('RA---TAN')
       call raDecToPixels(ra, dec, wcs_image, xs, ys)
       xpixel=xs
       ypixel=ys
    case('DETX')
       call raDecToDet(ra, dec, scAttitude, timeStamp, detx, dety)
       !write(*,*)'Det pix: ',detx, dety
       call linearToImPixels(detx, dety, xpixel, ypixel)
    case('DETX_ANG')
       call raDecToDet(ra, dec, scAttitude, timeStamp, detx, dety)
       !write(*,*)'Det pix: ',detx, dety
       call linearToImPixels(detx, dety, xpixel, ypixel)
    case('EQU--CAR') ! Flat, linear conversion to RA/DEC (degs) 
       call linearToImPixels(ra/PiOn180, dec/PiOn180, xpixel, ypixel)
    case default
       call error('notSupported',"This value of CTYPE1 is not yet supported.")
    end select

end subroutine callradectoimpixels

! Routine to convert between image pixels and ra/dec 
subroutine callimpixelstoradec(xpixel, ypixel, ra, dec)

    real(double), intent(in) :: xpixel, ypixel
    real(double), intent(out) :: ra, dec ! Output units are radians
    real(double) :: detx, dety
    real(single) :: xs, ys

   ! If the transform is RA---TAN then use the ssclib routine
    select case(wcs_image%x%ctype)
    case('RA---TAN')
       xs = xpixel ! ssclib routines needs single precision pixels
       ys = ypixel
       call pixelsToRaDec(xs, ys, wcs_image, ra, dec)
    case('EQU--CAR') ! Flat, linear conversion to RA/DEC (degs) 
       call imPixelsToLinear(xpixel, ypixel, ra, dec)
       ra = ra * PiOn180   ! Transform units most likely in degrees
       dec = dec * PiOn180
    case('DETX')
       call imPixelsToLinear(xpixel, ypixel, detx, dety)
       call detToRaDec(detx, dety, scAttitude, timeStamp, ra, dec)
    case('DETX_ANG')
       call imPixelsToLinear(xpixel, ypixel, detx, dety)
       call detToRaDec(detx, dety, scAttitude, timeStamp, ra, dec)
    case default
       call error('notSupported',"This value of CTYPE1 is not yet supported.")
    end select

end subroutine callimpixelstoradec

! Routine to convert between ra/dec and sky pixel
subroutine callradectoskypixels(ra, dec, xpixel, ypixel)

    real(double), intent(in) :: ra, dec
    real(double), intent(out) :: xpixel, ypixel
    real(single) :: xs, ys

    call raDecToPixels(ra, dec, wcs_sky, xs, ys)
    xpixel=xs
    ypixel=ys

end subroutine callradectoskypixels

! Routine to convert between detector pixels and ra/dec
subroutine calldettoradec(detx, dety, ra, dec)

    real(double), intent(in) :: detx, dety
    real(double), intent(out) :: ra, dec

    call detToRaDec(detx, dety, scAttitude, timeStamp, ra, dec)

end subroutine calldettoradec

! Routine to convert between detector pixels and theta/phi
subroutine calldettotel(detx, dety, theta, phi)

    real(double), intent(in) :: detx, dety
    real(double), intent(out) :: theta ! Arcseconds
    real(double), intent(out) :: phi ! Radians

    call detToTel(detx, dety, theta, phi)

end subroutine calldettotel

! Routine to convert between ra/dec and theta/phi
subroutine callradectotel(ra, dec, theta, phi)

    real(double), intent(in) :: ra, dec ! radians
    real(double), intent(out) :: theta ! Arcseconds
    real(double), intent(out) :: phi ! Radians

    call raDecToTel(ra, dec, scAttitude, timeStamp, theta, phi)

end subroutine callradectotel

! Routine to convert between ra/dec and detector pixels 
subroutine callradectodet(ra, dec, detx, dety)

    real(double), intent(in) :: ra, dec
    real(double), intent(out) :: detx, dety

    call raDecToDet(ra, dec, scAttitude, timeStamp, detx, dety)

end subroutine callradectodet


! Routine to convert between detector pixels and sky (x/y) coords
subroutine calldettoxy(detx, dety, x, y)

    real(double), intent(in) :: detx, dety
    real(double), intent(out) :: x, y
    real(double) :: ra, dec
    real(single) :: xs, ys

    call detToRaDec(detx, dety, scAttitude, timeStamp, ra, dec)
!    write(*,*)'RA conv: ',detx, dety,ra, dec
    call raDecToPixels(ra, dec, wcs_sky, xs, ys)
!    write(*,*)'RA conv2: ',ra,dec,xs,ys
    x=xs; y=ys;

end subroutine calldettoxy

! Routine to convert between detector pixels and raw coords
subroutine calldettoraw(detx, dety, rawx, rawy, ccd, isOffChip)

    real(double), intent(in) :: detx, dety
    real(double), intent(out) :: rawx, rawy
    integer(int8), intent(out) :: ccd
    !logical(bool), intent(out) :: isOffChip
    logical*1, intent(out) :: isOffChip

    call detToRaw(detx, dety, rawx, rawy, isOffChip, ccd)

end subroutine calldettoraw


! Routine to convert between sky (X/Y) and detector pixels
subroutine callxytodet(x, y, detx, dety)

    real(double), intent(in) :: x, y ! Sky coords
    real(double), intent(out) :: detx, dety
    real(double) :: ra,dec
    real(single) :: xpixel, ypixel

    xpixel = x 
    ypixel = y
    call pixelsToRaDec(xpixel, ypixel, wcs_sky, ra, dec)
    call raDecToDet(ra, dec, scAttitude, timeStamp, detx, dety)

end subroutine callxytodet

! Routine to convert between sky (X/Y) and detector pixels
! CCD number must have been set in CAL first
subroutine callrawtodet(rawx, rawy, detx, dety)

    real(double), intent(in) :: rawx, rawy ! Raw chip coordinates
    real(double), intent(out) :: detx, dety

    call rawToDet(rawx, rawy, detx, dety)

end subroutine callrawtodet

! Routine to convert between sky (X/Y) pixels and RA/DEC
subroutine callxytoradec(x, y, ra, dec)

    real(double), intent(in) :: x, y ! Sky coords
    real(double), intent(out) :: ra,dec
    real(single) :: xpixel, ypixel

    xpixel = x
    ypixel = y
    call pixelsToRaDec(xpixel, ypixel, wcs_sky, ra, dec)

end subroutine callxytoradec

subroutine callimpixelstodet(xpixel, ypixel, detx, dety)

    real(double), intent(in) :: xpixel, ypixel 
    real(double), intent(out) :: detx, dety

    call imPixelsToLinear(xpixel, ypixel, detx, dety)

end subroutine callimpixelstodet
!
! Routine to set the internal WCS structure to be used by the FORTRAN
! routines when converting between coordinate systems
!
subroutine setwcs(fileNameBytes, length)

    !character(*), intent(in) :: filename
    !character(20), intent(in) :: filename
    integer(kind=1), intent(in)  :: fileNameBytes(1024)
    integer, intent(in) :: length

    character(1024) filename
    type(DataSetT) :: set
    type(TableT) :: tab
    integer ftype
    integer i

    write (filename,'(1024a)') (achar(fileNameBytes(i)),i=1,length)

   ! Check the transformation - if it is not RA---TAN or DET then set
   ! wcs_image here
   ! call checkTransform(filename(1:length), wcs_image)

   ! If the first block is an array use it directly
    ftype = wcsfiletype(filename(1:length))

    if (ftype .eq. 0) then !ARRAY
       call readWcs(filename(1:length), wcs_image)
    else if (ftype .eq. 1) then !EVENTFILE
       set = dataSet(filename(1:length), READ)
       tab = table(set, 'EVENTS')
       call readWcs(tab, 'X', 'Y', wcs_image)
       call release(set)
    else
       call error('badWcsFile',"Can not read WCS keywords from this file.")
    endif

   ! If it is a DETX_ANG transformation need to convert from degrees to pixels
    if (wcs_image%x%ctype .eq. 'DETX_ANG') then ! Or check for units=degrees ?
       wcs_image%x%matrixOnAxis = wcs_image%x%matrixOnAxis * 72000.0
       wcs_image%y%matrixOnAxis = wcs_image%y%matrixOnAxis * 72000.0
       wcs_image%x%refWorld = wcs_image%x%refWorld * 72000.0
       wcs_image%y%refWorld = wcs_image%y%refWorld * 72000.0
    endif

end subroutine setwcs
    
!
! Routine to set the internal WCS structure to be used by the FORTRAN
! routines when converting between coordinate systems (REF keywords)
! This works for images or event tables
!
subroutine setwcsref(fileNameBytes, length)

    integer(kind=1), intent(in)  :: fileNameBytes(1024)
    integer, intent(in) :: length

    character(1024) filename
    integer i

    write (filename,'(1024a)') (achar(fileNameBytes(i)),i=1,length)

    call readWcsSkyKeys(filename(1:length), wcs_sky)

!     write(*,*)' WCS sky: ',wcs_sky
end subroutine setwcsref
!
! Routine to set the internal attitude structure to be used by the FORTRAN
! routines when converting between coordinate systems
!
subroutine setatt(fileNameBytes, length)

    integer(kind=1), intent(in)  :: fileNameBytes(1024)
    integer, intent(in) :: length

    character(1024) filename
    integer i

    type(DataSetT) :: set

    write (filename,'(1024a)') (achar(fileNameBytes(i)),i=1,length)

    ! This should be the call but causes seg faults - do it piecewise
    !call getCalInfo(filename(1:length), instrumentId,& 
     !          timeStamp, scAttitude, instrumAttitude)

    set = dataSet(filename(1:length), READ)

    ! Set the spacecraft attitude
    scAttitude%ra = real64Attribute(set, "RA_PNT") * PiOn180
    scAttitude%dec  = real64Attribute(set, "DEC_PNT") * PiOn180
    scAttitude%apos = real64Attribute(set, "PA_PNT") * PiOn180

    ! Set the observation time
    timeStamp = toTimeTag(stringAttribute(set, "DATE-OBS"))
    
    ! Get the instrument attitude after setting the instrument in the CAL
    instrumentId = instId(stringAttribute(set, "INSTRUME"))

    call CAL_setState(instrument=instrumentId, randomize=.false.)
    instrumAttitude = getInstrumentAttitude(scAttitude, timeStamp)

    !write(*,*)' SC att:',scAttitude
    !write(*,*)' inst att:',instrumAttitude
    !write(*,*)' timestamp:',timeStamp

    ! Release this reference to the dataset
!    call release(set)

end subroutine setatt

!
! Routine to convert between a linear coordinate system and image pixels
! i.e. not to be used with the tangential projection
!
subroutine linearToImPixels(xin, yin, xpixel, ypixel)

    real(double), intent(in) :: xin, yin   ! Coordinates to transform
    real(double), intent(out) :: xpixel, ypixel
    real(double) :: rx, ry, lx, ly

    ! Use the wcs_image construct to make a linear transormation between
    ! the input and output coordinates
    rx = xin - wcs_image%x%refWorld
    ry = yin - wcs_image%y%refWorld
    lx = rx / wcs_image%x%matrixOnAxis
    ly = ry / wcs_image%y%matrixOnAxis
    
    !write(*,*)'RA,DEC input:',xin, yin
    !write(*,*)'Nums ',rx,ry,lx,ly
    !write(*,*)'refPix refWorld: ',wcs_image%x%refPixels,wcs_image%x%refWorld,&
    !              wcs_image%x%matrixOnAxis
    xpixel = lx + wcs_image%x%refPixels
    ypixel = ly + wcs_image%y%refPixels

end subroutine linearToImPixels

!
! Routine to convert between image pixels to a linear coordinate system 
! i.e. not to be used with the tangential projection
!
subroutine imPixelsToLinear(xpixel, ypixel, xout, yout)

    real(double), intent(in) :: xpixel, ypixel
    real(double), intent(out) :: xout, yout   ! Coordinates to transform
    real(double) :: rx, ry, lx, ly

    ! Use the wcs_image construct to make a linear transormation between
    ! the input and output coordinates
    rx = xpixel - wcs_image%x%refPixels
    ry = ypixel - wcs_image%y%refPixels
    lx = rx * wcs_image%x%matrixOnAxis
    ly = ry * wcs_image%y%matrixOnAxis
    
    !write(*,*)'Input pixel:',xpixel, ypixel
    !write(*,*)'Nums ',rx,ry,lx,ly
    !write(*,*)'refPix refWorld: ',wcs_image%x%refPixels,wcs_image%x%refWorld,&
    !              wcs_image%x%matrixOnAxis
    xout = lx + wcs_image%x%refWorld
    yout = ly + wcs_image%y%refWorld
    !write(*,*)'Output:',xout,yout

end subroutine imPixelsToLinear

!
! Check if a file contains an array in the first block (i.e. an image)
!
function wcsfiletype(filename) result(wcsfileresult)

    implicit none
    character(*)    :: filename
    integer    :: wcsfileresult
    type(DataSetT) :: set

    ! Open the file and check if the first block is an array
    ! or if it is an event file
    set = dataSet(filename, READ)
    if ( blockType( set, 0 ) == ARRAY_BLOCK ) then
       wcsfileresult=0 !'ARRAY'
    else if (name(block(set, 0)) .eq. 'EVENTS') then  
       wcsfileresult=1 !'EVENTFILE'
    else
       wcsfileresult=-1 !'NotKnown'
    endif

    call release(set)
    
end function wcsfiletype
    
end module ssclib_glue

